# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictBool, StrictStr
from client.models.env_var import EnvVar
from client.models.resource_request import ResourceRequest
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class Transformer(BaseModel):
    """
    Transformer
    """ # noqa: E501
    id: Optional[StrictStr] = None
    enabled: Optional[StrictBool] = None
    transformer_type: Optional[StrictStr] = None
    image: Optional[StrictStr] = None
    command: Optional[StrictStr] = None
    args: Optional[StrictStr] = None
    resource_request: Optional[ResourceRequest] = None
    env_vars: Optional[List[EnvVar]] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    __properties: ClassVar[List[str]] = ["id", "enabled", "transformer_type", "image", "command", "args", "resource_request", "env_vars", "created_at", "updated_at"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of Transformer from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of resource_request
        if self.resource_request:
            _dict['resource_request'] = self.resource_request.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in env_vars (list)
        _items = []
        if self.env_vars:
            for _item in self.env_vars:
                if _item:
                    _items.append(_item.to_dict())
            _dict['env_vars'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of Transformer from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "id": obj.get("id"),
            "enabled": obj.get("enabled"),
            "transformer_type": obj.get("transformer_type"),
            "image": obj.get("image"),
            "command": obj.get("command"),
            "args": obj.get("args"),
            "resource_request": ResourceRequest.from_dict(obj.get("resource_request")) if obj.get("resource_request") is not None else None,
            "env_vars": [EnvVar.from_dict(_item) for _item in obj.get("env_vars")] if obj.get("env_vars") is not None else None,
            "created_at": obj.get("created_at"),
            "updated_at": obj.get("updated_at")
        })
        return _obj





from __future__ import annotations

import client
from typing import Optional
from merlin.util import autostr
from dataclasses import dataclass
from dataclasses_json import dataclass_json


@autostr
@dataclass_json
@dataclass
class GroundTruthSource:
    table_urn: str
    event_timestamp_column: str
    source_project: str


@autostr
@dataclass_json
@dataclass
class GroundTruthJob:
    cron_schedule: str
    service_account_secret_name: str
    start_day_offset_from_now: int
    end_day_offset_from_now: int
    cpu_request: Optional[str] = None
    cpu_limit: Optional[str] = None
    memory_request: Optional[str] = None
    memory_limit: Optional[str] = None
    grace_period_day: Optional[int] = None


@autostr
@dataclass_json
@dataclass
class PredictionLogIngestionResourceRequest:
    replica: int
    cpu_request: Optional[str] = None
    memory_request: Optional[str] = None


@autostr
@dataclass_json
@dataclass
class ModelObservability:
    enabled: bool
    ground_truth_source: Optional[GroundTruthSource] = None
    ground_truth_job: Optional[GroundTruthJob] = None
    prediction_log_ingestion_resource_request: Optional[PredictionLogIngestionResourceRequest] = None

    def to_model_observability_request(self) -> client.ModelObservability:
        ground_truth_source = None
        if self.ground_truth_source is not None:
            ground_truth_source = client.GroundTruthSource(
                table_urn=self.ground_truth_source.table_urn,
                event_timestamp_column=self.ground_truth_source.event_timestamp_column,
                source_project=self.ground_truth_source.source_project
            )

        ground_truth_job = None
        if self.ground_truth_job is not None:
            ground_truth_job = client.GroundTruthJob(
                cron_schedule=self.ground_truth_job.cron_schedule,
                service_account_secret_name=self.ground_truth_job.service_account_secret_name,
                start_day_offset_from_now=self.ground_truth_job.start_day_offset_from_now,
                end_day_offset_from_now=self.ground_truth_job.end_day_offset_from_now,
                cpu_request=self.ground_truth_job.cpu_request,
                cpu_limit=self.ground_truth_job.cpu_limit,
                memory_request=self.ground_truth_job.memory_request,
                memory_limit=self.ground_truth_job.memory_limit,
                grace_period_day=self.ground_truth_job.grace_period_day
            )
        
        prediction_log_ingestion_resource_request = None
        if self.prediction_log_ingestion_resource_request is not None:
            prediction_log_ingestion_resource_request = client.PredictionLogIngestionResourceRequest(
                replica=self.prediction_log_ingestion_resource_request.replica,
                cpu_request=self.prediction_log_ingestion_resource_request.cpu_request,
                memory_request=self.prediction_log_ingestion_resource_request.memory_request
            )

        return client.ModelObservability(
            enabled=self.enabled,
            ground_truth_source=ground_truth_source,
            ground_truth_job=ground_truth_job,
            prediction_log_ingestion_resource_request=prediction_log_ingestion_resource_request
        )

    @classmethod
    def from_model_observability_response(cls, response: Optional[client.ModelObservability]) -> Optional[ModelObservability]:
        if response is None:
            return None

        ground_truth_source = None
        if response.ground_truth_source is not None:
            ground_truth_source = cls._ground_truth_source_from_response(response.ground_truth_source)

        ground_truth_job = None
        if response.ground_truth_job is not None:
            ground_truth_job = cls._job_from_response(response.ground_truth_job)

        prediction_log_ingestion_resource_request = None
        if response.prediction_log_ingestion_resource_request is not None:
            prediction_log_ingestion_resource_request = cls._prediction_log_ingestion_resource_request_from_response(response.prediction_log_ingestion_resource_request)

        return ModelObservability(
            enabled=response.enabled,
            ground_truth_source=ground_truth_source,
            ground_truth_job=ground_truth_job,
            prediction_log_ingestion_resource_request=prediction_log_ingestion_resource_request
        )
    
    @classmethod
    def _ground_truth_source_from_response(cls, response: Optional[client.GroundTruthSource]) -> Optional[GroundTruthSource]:
        if response is None:
            return None

        return GroundTruthSource(
            table_urn=response.table_urn,
            event_timestamp_column=response.event_timestamp_column,
            source_project=response.source_project
        )
    
    @classmethod
    def _job_from_response(cls, response: Optional[client.GroundTruthJob]) -> Optional[GroundTruthJob]:
        if response is None:
            return None

        return GroundTruthJob(
            cron_schedule=response.cron_schedule,
            service_account_secret_name=response.service_account_secret_name,
            start_day_offset_from_now=response.start_day_offset_from_now,
            end_day_offset_from_now=response.end_day_offset_from_now,
            cpu_request=response.cpu_request,
            cpu_limit=response.cpu_limit,
            memory_request=response.memory_request,
            memory_limit=response.memory_limit,
            grace_period_day=response.grace_period_day
        )
    
    @classmethod
    def _prediction_log_ingestion_resource_request_from_response(cls, response: Optional[client.PredictionLogIngestionResourceRequest]) -> Optional[PredictionLogIngestionResourceRequest]:
        if response is None:
            return None

        return PredictionLogIngestionResourceRequest(
            replica=response.replica,
            cpu_request=response.cpu_request,
            memory_request=response.memory_request
        )


from enum import Enum
from typing import Optional

import client
from merlin.util import autostr


class ImageBuildingJobState(Enum):
    ACTIVE = "active"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    UNKNOWN = "unknown"


class ImageBuildingJobStatus:
    def __init__(self, status: client.ImageBuildingJobStatus):
        self._state = ImageBuildingJobState(status.state)
        self._message = status.message

    @property
    def state(self) -> Optional[ImageBuildingJobState]:
        return self._state

    @property
    def message(self) -> Optional[str]:
        return self._message


@autostr
class VersionImage:
    def __init__(self, image: client.VersionImage):
        self._project_id = image.project_id
        self._model_id = image.model_id
        self._version_id = image.version_id
        self._image_ref = image.image_ref
        self._exists = image.exists
        if image.image_building_job_status is not None:
            self._image_building_job_status = ImageBuildingJobStatus(
                image.image_building_job_status
            )

    @property
    def project_id(self) -> Optional[int]:
        return self._project_id

    @property
    def model_id(self) -> Optional[int]:
        return self._model_id

    @property
    def version_id(self) -> Optional[int]:
        return self._version_id

    @property
    def image_ref(self) -> Optional[str]:
        return self._image_ref

    @property
    def exists(self) -> Optional[bool]:
        return self._exists

    @property
    def image_building_job_status(self) -> Optional[ImageBuildingJobStatus]:
        return self._image_building_job_status



import os
import uuid
from test.utils import undeploy_all_version
from time import sleep
from typing import List

import grpc
import pandas as pd
import pytest
import xgboost as xgb
from caraml.upi.utils import df_to_table, table_to_df
from caraml.upi.v1 import type_pb2, upi_pb2, upi_pb2_grpc, variable_pb2
from google.protobuf.json_format import MessageToDict, ParseDict
from merlin.deployment_mode import DeploymentMode
from merlin.endpoint import Status
from merlin.model import ModelType, PyFuncModel
from merlin.protocol import Protocol
from merlin.transformer import StandardTransformer
from sklearn.datasets import load_iris

import merlin


class IrisClassifier(PyFuncModel):
    target_names = ["setosa", "versicolor", "virginica"]

    target_name = "iris-species"

    def initialize(self, artifacts):
        self._model = xgb.Booster(model_file=artifacts["xgb_model"])

    def infer(self, request: dict, **kwargs):
        result = self._predict(request["instances"])
        return {"predictions": result.tolist()}

    def upiv1_infer(
        self, request: upi_pb2.PredictValuesRequest, context: grpc.ServicerContext
    ) -> upi_pb2.PredictValuesResponse:
        features_df, _ = table_to_df(request.prediction_table)
        prediction_result_df = self._predict(features_df)
        return self._create_response(prediction_result_df, request)

    def _create_response(
        self, predictions: pd.DataFrame, request: upi_pb2.PredictValuesRequest
    ) -> upi_pb2.PredictValuesResponse:
        prediction_result_table = df_to_table(predictions, "prediction_result")
        response_metadata = upi_pb2.ResponseMetadata(
            prediction_id=request.metadata.prediction_id
        )
        return upi_pb2.PredictValuesResponse(
            prediction_result_table=prediction_result_table,
            target_name=self.target_name,
            metadata=response_metadata,
        )

    def _predict(self, features: pd.DataFrame) -> pd.DataFrame:
        features_matrix = xgb.DMatrix(features)
        return pd.DataFrame(
            self._model.predict(features_matrix), columns=self.target_names
        )


class SimpleForwarder(PyFuncModel):
    target_name = "probability"

    def infer(self, request: dict, **kwargs):
        upi_request = upi_pb2.PredictValuesRequest()
        ParseDict(request, upi_request)
        upi_response = upi_pb2.PredictValuesResponse(
            prediction_result_table=upi_request.prediction_table,
            target_name=self.target_name,
        )
        return MessageToDict(upi_response)

    def upiv1_infer(
        self, request: upi_pb2.PredictValuesRequest, context: grpc.ServicerContext
    ) -> upi_pb2.PredictValuesResponse:
        return upi_pb2.PredictValuesResponse(
            prediction_result_table=request.prediction_table,
            target_name=self.target_name,
        )


@pytest.mark.pyfunc
@pytest.mark.integration
def test_deploy(integration_test_url, project_name, use_google_oauth, requests):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pyfunc-upi", ModelType.PYFUNC)

    undeploy_all_version()
    with merlin.new_model_version() as v:
        xgb_path, xgb_model = train_xgboost_model()

        v.log_pyfunc_model(
            model_instance=IrisClassifier(),
            conda_env="test/pyfunc/env.yaml",
            code_dir=["test"],
            artifacts={"xgb_model": xgb_path},
        )

    endpoint = merlin.deploy(v, protocol=Protocol.UPI_V1)

    assert endpoint.protocol == Protocol.UPI_V1
    assert endpoint.status == Status.RUNNING
    assert endpoint.deployment_mode == DeploymentMode.SERVERLESS

    channel = grpc.insecure_channel(f"{endpoint.url}:80")
    stub = upi_pb2_grpc.UniversalPredictionServiceStub(channel)

    validate_iris_upi(xgb_model, stub)
    merlin.undeploy(v)


@pytest.mark.pyfunc
@pytest.mark.integration
@pytest.mark.parametrize("upi_http_predictor", [False, True])
def test_pyfunc_with_standard_transformer(
    integration_test_url, project_name, upi_http_predictor, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    model_name = "pyfunc-upi-std-http" if upi_http_predictor else "pyfunc-upi-std"
    merlin.set_model(model_name, ModelType.PYFUNC)

    undeploy_all_version()
    with merlin.new_model_version() as v:
        v.log_pyfunc_model(
            model_instance=SimpleForwarder(),
            conda_env="test/pyfunc/env.yaml",
            code_dir=["test"],
        )

    transformer_config_path = os.path.join(
        "test/transformer", "upi_standard_transformer_no_feast.yaml"
    )

    upi_http_enabled_val = "true" if upi_http_predictor else "false"
    transformer = StandardTransformer(
        config_file=transformer_config_path,
        enabled=True,
        env_vars={"PREDICTOR_UPI_HTTP_ENABLED": upi_http_enabled_val},
    )
    endpoint = merlin.deploy(v, transformer=transformer, protocol=Protocol.UPI_V1)

    assert endpoint.protocol == Protocol.UPI_V1
    assert endpoint.status == Status.RUNNING

    channel = grpc.insecure_channel(f"{endpoint.url}:80")
    stub = upi_pb2_grpc.UniversalPredictionServiceStub(channel)

    print(endpoint.url)
    sleep(5)

    request = create_simple_forwarder_request()
    response = stub.PredictValues(request=request)

    assert response.metadata.prediction_id == request.metadata.prediction_id
    assert response.target_name == request.target_name
    exp_response = simple_forwarder_response()
    assert response == exp_response

    merlin.undeploy(v)


@pytest.mark.pyfunc
@pytest.mark.integration
def test_serve_traffic(integration_test_url, project_name, use_google_oauth, requests):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pyfunc-upi-serve", ModelType.PYFUNC)

    undeploy_all_version()
    with merlin.new_model_version() as v:
        xgb_path, xgb_model = train_xgboost_model()

        v.log_pyfunc_model(
            model_instance=IrisClassifier(),
            conda_env="test/pyfunc/env.yaml",
            code_dir=["test"],
            artifacts={"xgb_model": xgb_path},
        )

    endpoint = merlin.deploy(v, protocol=Protocol.UPI_V1)
    model_endpoint = merlin.serve_traffic({endpoint: 100})

    assert model_endpoint.protocol == Protocol.UPI_V1
    assert model_endpoint.status == Status.SERVING

    channel = grpc.insecure_channel(f"{model_endpoint.url}:80")
    stub = upi_pb2_grpc.UniversalPredictionServiceStub(channel)

    print(model_endpoint.url)
    sleep(5)
    validate_iris_upi(xgb_model, stub)

    merlin.stop_serving_traffic(model_endpoint.environment_name)
    merlin.undeploy(v)


def test_model():
    model_path, xgb_model = train_xgboost_model()
    pyfunc_model = IrisClassifier()
    pyfunc_model.initialize({"xgb_model": model_path})
    request = create_upi_request_from_iris_dataset()
    response = pyfunc_model.upiv1_infer(request, {})

    X = load_iris()["data"]
    y = xgb_model.predict(xgb.DMatrix(X))
    exp_df = pd.DataFrame(y, columns=IrisClassifier.target_names)
    exp_table = df_to_table(exp_df, "prediction_result")

    assert exp_table == response.prediction_result_table


def validate_iris_upi(model, stub):
    request = create_upi_request_from_iris_dataset()
    response = stub.PredictValues(request=request)

    assert response.metadata.prediction_id == request.metadata.prediction_id
    assert response.target_name == request.target_name

    X = load_iris()["data"]
    y = model.predict(xgb.DMatrix(X))
    exp_df = pd.DataFrame(y, columns=IrisClassifier.target_names)
    exp_table = df_to_table(exp_df, "prediction_result")

    assert exp_table == response.prediction_result_table


def train_xgboost_model():
    model_dir = "test/pyfunc/"
    BST_FILE = "model_1.bst"

    iris = load_iris()
    y = iris["target"]
    X = iris["data"]

    dtrain = xgb.DMatrix(X, label=y)
    param = {
        "max_depth": 6,
        "eta": 0.1,
        "silent": 1,
        "nthread": 4,
        "num_class": 3,
        "objective": "multi:softprob",
    }
    xgb_model = xgb.train(params=param, dtrain=dtrain)
    model_path = os.path.join(model_dir, BST_FILE)
    xgb_model.save_model(model_path)
    return model_path, xgb_model


def create_simple_forwarder_request() -> upi_pb2.PredictValuesRequest:
    target_name = SimpleForwarder.target_name
    cols = [
        "id",
        "name",
        "vehicle",
        "previous_vehicle",
        "rating",
        "test_time",
        "row_number",
    ]
    indices = ["row1", "row2"]
    rows = [
        [1, "driver-1", "motorcycle", "suv", 4.0, 90, 0],
        [2, "driver-2", "sedan", "mpv", 3.0, 90, 1],
    ]
    driver_df = pd.DataFrame(columns=cols, data=rows, index=indices)
    driver_table = df_to_table(driver_df, "driver_table")
    variables: List[variable_pb2.Variable] = [
        variable_pb2.Variable(
            name="customer_id", type=type_pb2.TYPE_INTEGER, integer_value=1111
        )
    ]
    return upi_pb2.PredictValuesRequest(
        target_name=target_name,
        transformer_input=upi_pb2.TransformerInput(
            tables=[driver_table], variables=variables
        ),
    )


def simple_forwarder_response() -> upi_pb2.PredictValuesResponse:
    target_name = SimpleForwarder.target_name
    cols = ["customer_id", "name", "rank", "rating", "vehicle", "previous_vehicle"]
    indices = ["row2", "row1"]
    rows = [[1111, "driver-2", 2.5, 0.5, 2, 3], [1111, "driver-1", -2.5, 0.75, 0, 1]]
    response_df = pd.DataFrame(columns=cols, data=rows, index=indices)
    response_table = df_to_table(response_df, "transformed_driver_table")
    return upi_pb2.PredictValuesResponse(
        target_name=target_name, prediction_result_table=response_table
    )


def create_upi_request_from_iris_dataset() -> upi_pb2.PredictValuesRequest:
    target_name = IrisClassifier.target_name
    iris_dataset = load_iris()
    X = iris_dataset["data"]
    df = pd.DataFrame(X, columns=iris_dataset.feature_names)

    prediction_table = df_to_table(df, "features")
    return upi_pb2.PredictValuesRequest(
        target_name=target_name,
        prediction_table=prediction_table,
        metadata=upi_pb2.RequestMetadata(prediction_id=str(uuid.uuid1())),
    )



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import io
import warnings

from pydantic import validate_call, Field, StrictFloat, StrictStr, StrictInt
from typing import Dict, List, Optional, Tuple, Union, Any

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

from pydantic import Field
from typing_extensions import Annotated
from pydantic import StrictInt, StrictStr

from typing import List, Optional

from client.models.version import Version

from client.api_client import ApiClient
from client.api_response import ApiResponse
from client.rest import RESTResponseType


class VersionApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client


    @validate_call
    def models_model_id_versions_get(
        self,
        model_id: StrictInt,
        limit: Optional[StrictInt] = None,
        cursor: Optional[StrictStr] = None,
        search: Annotated[Optional[StrictStr], Field(description="Search query to filter the model versions. These searches are currently supported: - Search by \"mlflow_run_id\" e.g. `?search=cfca7716b45f4b149479630a98332a13` - Search by \"environment_name\" e.g `?search=environment_name:myenv` - Search by \"labels\" e.g. `?search=labels:app IN (nginx,postgres), country in (SG)` - Search by \"environment_name\" and \"labels\" e.g.   `?search=environment_name:myenv labels:app IN (nginx,postgres), country in (SG)` ")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> List[Version]:
        """List versions of the models


        :param model_id: (required)
        :type model_id: int
        :param limit:
        :type limit: int
        :param cursor:
        :type cursor: str
        :param search: Search query to filter the model versions. These searches are currently supported: - Search by \"mlflow_run_id\" e.g. `?search=cfca7716b45f4b149479630a98332a13` - Search by \"environment_name\" e.g `?search=environment_name:myenv` - Search by \"labels\" e.g. `?search=labels:app IN (nginx,postgres), country in (SG)` - Search by \"environment_name\" and \"labels\" e.g.   `?search=environment_name:myenv labels:app IN (nginx,postgres), country in (SG)` 
        :type search: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_get_serialize(
            model_id=model_id,
            limit=limit,
            cursor=cursor,
            search=search,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[Version]",
            '404': None
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def models_model_id_versions_get_with_http_info(
        self,
        model_id: StrictInt,
        limit: Optional[StrictInt] = None,
        cursor: Optional[StrictStr] = None,
        search: Annotated[Optional[StrictStr], Field(description="Search query to filter the model versions. These searches are currently supported: - Search by \"mlflow_run_id\" e.g. `?search=cfca7716b45f4b149479630a98332a13` - Search by \"environment_name\" e.g `?search=environment_name:myenv` - Search by \"labels\" e.g. `?search=labels:app IN (nginx,postgres), country in (SG)` - Search by \"environment_name\" and \"labels\" e.g.   `?search=environment_name:myenv labels:app IN (nginx,postgres), country in (SG)` ")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[List[Version]]:
        """List versions of the models


        :param model_id: (required)
        :type model_id: int
        :param limit:
        :type limit: int
        :param cursor:
        :type cursor: str
        :param search: Search query to filter the model versions. These searches are currently supported: - Search by \"mlflow_run_id\" e.g. `?search=cfca7716b45f4b149479630a98332a13` - Search by \"environment_name\" e.g `?search=environment_name:myenv` - Search by \"labels\" e.g. `?search=labels:app IN (nginx,postgres), country in (SG)` - Search by \"environment_name\" and \"labels\" e.g.   `?search=environment_name:myenv labels:app IN (nginx,postgres), country in (SG)` 
        :type search: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_get_serialize(
            model_id=model_id,
            limit=limit,
            cursor=cursor,
            search=search,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[Version]",
            '404': None
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def models_model_id_versions_get_without_preload_content(
        self,
        model_id: StrictInt,
        limit: Optional[StrictInt] = None,
        cursor: Optional[StrictStr] = None,
        search: Annotated[Optional[StrictStr], Field(description="Search query to filter the model versions. These searches are currently supported: - Search by \"mlflow_run_id\" e.g. `?search=cfca7716b45f4b149479630a98332a13` - Search by \"environment_name\" e.g `?search=environment_name:myenv` - Search by \"labels\" e.g. `?search=labels:app IN (nginx,postgres), country in (SG)` - Search by \"environment_name\" and \"labels\" e.g.   `?search=environment_name:myenv labels:app IN (nginx,postgres), country in (SG)` ")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """List versions of the models


        :param model_id: (required)
        :type model_id: int
        :param limit:
        :type limit: int
        :param cursor:
        :type cursor: str
        :param search: Search query to filter the model versions. These searches are currently supported: - Search by \"mlflow_run_id\" e.g. `?search=cfca7716b45f4b149479630a98332a13` - Search by \"environment_name\" e.g `?search=environment_name:myenv` - Search by \"labels\" e.g. `?search=labels:app IN (nginx,postgres), country in (SG)` - Search by \"environment_name\" and \"labels\" e.g.   `?search=environment_name:myenv labels:app IN (nginx,postgres), country in (SG)` 
        :type search: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_get_serialize(
            model_id=model_id,
            limit=limit,
            cursor=cursor,
            search=search,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[Version]",
            '404': None
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _models_model_id_versions_get_serialize(
        self,
        model_id,
        limit,
        cursor,
        search,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> Tuple:

        _host = None

        _collection_formats: Dict[str, str] = {
            
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, str] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if model_id is not None:
            _path_params['model_id'] = model_id
        # process the query parameters
        if limit is not None:
            
            _query_params.append(('limit', limit))
            
        if cursor is not None:
            
            _query_params.append(('cursor', cursor))
            
        if search is not None:
            
            _query_params.append(('search', search))
            
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            [
                '*/*'
            ]
        )


        # authentication setting
        _auth_settings: List[str] = [
            'Bearer'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/models/{model_id}/versions',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    def models_model_id_versions_post(
        self,
        model_id: StrictInt,
        body: Optional[Version] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> Version:
        """Log a new version of the models


        :param model_id: (required)
        :type model_id: int
        :param body:
        :type body: Version
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_post_serialize(
            model_id=model_id,
            body=body,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "Version"
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def models_model_id_versions_post_with_http_info(
        self,
        model_id: StrictInt,
        body: Optional[Version] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[Version]:
        """Log a new version of the models


        :param model_id: (required)
        :type model_id: int
        :param body:
        :type body: Version
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_post_serialize(
            model_id=model_id,
            body=body,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "Version"
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def models_model_id_versions_post_without_preload_content(
        self,
        model_id: StrictInt,
        body: Optional[Version] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Log a new version of the models


        :param model_id: (required)
        :type model_id: int
        :param body:
        :type body: Version
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_post_serialize(
            model_id=model_id,
            body=body,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "Version"
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _models_model_id_versions_post_serialize(
        self,
        model_id,
        body,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> Tuple:

        _host = None

        _collection_formats: Dict[str, str] = {
            
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, str] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if model_id is not None:
            _path_params['model_id'] = model_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if body is not None:
            _body_params = body


        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            [
                '*/*'
            ]
        )


        # authentication setting
        _auth_settings: List[str] = [
            'Bearer'
        ]

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/models/{model_id}/versions',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    def models_model_id_versions_version_id_delete(
        self,
        model_id: StrictInt,
        version_id: StrictInt,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> int:
        """Delete version by ID from model


        :param model_id: (required)
        :type model_id: int
        :param version_id: (required)
        :type version_id: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_version_id_delete_serialize(
            model_id=model_id,
            version_id=version_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "int",
            '400': None,
            '404': None,
            '500': None
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def models_model_id_versions_version_id_delete_with_http_info(
        self,
        model_id: StrictInt,
        version_id: StrictInt,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[int]:
        """Delete version by ID from model


        :param model_id: (required)
        :type model_id: int
        :param version_id: (required)
        :type version_id: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_version_id_delete_serialize(
            model_id=model_id,
            version_id=version_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "int",
            '400': None,
            '404': None,
            '500': None
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def models_model_id_versions_version_id_delete_without_preload_content(
        self,
        model_id: StrictInt,
        version_id: StrictInt,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Delete version by ID from model


        :param model_id: (required)
        :type model_id: int
        :param version_id: (required)
        :type version_id: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_version_id_delete_serialize(
            model_id=model_id,
            version_id=version_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "int",
            '400': None,
            '404': None,
            '500': None
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _models_model_id_versions_version_id_delete_serialize(
        self,
        model_id,
        version_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> Tuple:

        _host = None

        _collection_formats: Dict[str, str] = {
            
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, str] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if model_id is not None:
            _path_params['model_id'] = model_id
        if version_id is not None:
            _path_params['version_id'] = version_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            [
                '*/*'
            ]
        )


        # authentication setting
        _auth_settings: List[str] = [
            'Bearer'
        ]

        return self.api_client.param_serialize(
            method='DELETE',
            resource_path='/models/{model_id}/versions/{version_id}',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    def models_model_id_versions_version_id_get(
        self,
        model_id: StrictInt,
        version_id: StrictInt,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> Version:
        """Get version by ID from model


        :param model_id: (required)
        :type model_id: int
        :param version_id: (required)
        :type version_id: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_version_id_get_serialize(
            model_id=model_id,
            version_id=version_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Version",
            '404': None
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def models_model_id_versions_version_id_get_with_http_info(
        self,
        model_id: StrictInt,
        version_id: StrictInt,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[Version]:
        """Get version by ID from model


        :param model_id: (required)
        :type model_id: int
        :param version_id: (required)
        :type version_id: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_version_id_get_serialize(
            model_id=model_id,
            version_id=version_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Version",
            '404': None
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def models_model_id_versions_version_id_get_without_preload_content(
        self,
        model_id: StrictInt,
        version_id: StrictInt,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get version by ID from model


        :param model_id: (required)
        :type model_id: int
        :param version_id: (required)
        :type version_id: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_version_id_get_serialize(
            model_id=model_id,
            version_id=version_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Version",
            '404': None
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _models_model_id_versions_version_id_get_serialize(
        self,
        model_id,
        version_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> Tuple:

        _host = None

        _collection_formats: Dict[str, str] = {
            
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, str] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if model_id is not None:
            _path_params['model_id'] = model_id
        if version_id is not None:
            _path_params['version_id'] = version_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            [
                '*/*'
            ]
        )


        # authentication setting
        _auth_settings: List[str] = [
            'Bearer'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/models/{model_id}/versions/{version_id}',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    def models_model_id_versions_version_id_patch(
        self,
        model_id: StrictInt,
        version_id: StrictInt,
        body: Optional[Version] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> None:
        """Patch the version 


        :param model_id: (required)
        :type model_id: int
        :param version_id: (required)
        :type version_id: int
        :param body:
        :type body: Version
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_version_id_patch_serialize(
            model_id=model_id,
            version_id=version_id,
            body=body,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def models_model_id_versions_version_id_patch_with_http_info(
        self,
        model_id: StrictInt,
        version_id: StrictInt,
        body: Optional[Version] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[None]:
        """Patch the version 


        :param model_id: (required)
        :type model_id: int
        :param version_id: (required)
        :type version_id: int
        :param body:
        :type body: Version
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_version_id_patch_serialize(
            model_id=model_id,
            version_id=version_id,
            body=body,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def models_model_id_versions_version_id_patch_without_preload_content(
        self,
        model_id: StrictInt,
        version_id: StrictInt,
        body: Optional[Version] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Patch the version 


        :param model_id: (required)
        :type model_id: int
        :param version_id: (required)
        :type version_id: int
        :param body:
        :type body: Version
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._models_model_id_versions_version_id_patch_serialize(
            model_id=model_id,
            version_id=version_id,
            body=body,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _models_model_id_versions_version_id_patch_serialize(
        self,
        model_id,
        version_id,
        body,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> Tuple:

        _host = None

        _collection_formats: Dict[str, str] = {
            
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, str] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if model_id is not None:
            _path_params['model_id'] = model_id
        if version_id is not None:
            _path_params['version_id'] = version_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if body is not None:
            _body_params = body




        # authentication setting
        _auth_settings: List[str] = [
            'Bearer'
        ]

        return self.api_client.param_serialize(
            method='PATCH',
            resource_path='/models/{model_id}/versions/{version_id}',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )





import pytest
import yaml
from merlin.requirements import process_conda_env


@pytest.mark.parametrize(
    "input,output,additional_merlin_reqs",
    [
        # Test using file
        (
            "test/requirements/empty_in.yaml",
            "test/requirements/empty_out.yaml",
            ["merlin-pyfunc-server<0.42.0"],
        ),
        (
            "test/requirements/no_pip_reqs_in.yaml",
            "test/requirements/no_pip_reqs_out.yaml",
            ["merlin-pyfunc-server<0.42.0"],
        ),
        (
            "test/requirements/pyfunc_server_with_version_in.yaml",
            "test/requirements/pyfunc_server_with_version_out.yaml",
            ["merlin-pyfunc-server<0.42.0"],
        ),
        (
            "test/requirements/pyfunc_server_without_version_in.yaml",
            "test/requirements/pyfunc_server_without_version_out.yaml",
            ["merlin-pyfunc-server<0.42.0"],
        ),
        (
            "test/requirements/non_package_reqs_in.yaml",
            "test/requirements/non_package_reqs_out.yaml",
            ["merlin-pyfunc-server<0.42.0"],
        ),
        (
            "test/requirements/other_reqs_in.yaml",
            "test/requirements/other_reqs_out.yaml",
            ["merlin-pyfunc-server<0.42.0"],
        ),
        # Conda files have pip dependencies that use requirements.txt
        (
            "test/requirements/with-requirements-txt/empty_in.yaml",
            "test/requirements/with-requirements-txt/empty_out.yaml",
            ["merlin-pyfunc-server<0.42.0"],
        ),
        (
            "test/requirements/with-requirements-txt/no_pyfunc_server_in.yaml",
            "test/requirements/with-requirements-txt/no_pyfunc_server_out.yaml",
            ["merlin-pyfunc-server<0.42.0"],
        ),
        (
            "test/requirements/with-requirements-txt/pyfunc_server_with_version_in.yaml",
            "test/requirements/with-requirements-txt/pyfunc_server_with_version_out.yaml",
            ["merlin-pyfunc-server<0.42.0"],
        ),
        (
            "test/requirements/with-requirements-txt/pyfunc_server_without_version_in.yaml",
            "test/requirements/with-requirements-txt/pyfunc_server_without_version_out.yaml",
            ["merlin-pyfunc-server<0.42.0"],
        ),
        (
            "test/requirements/with-requirements-txt/constraints_in.yaml",
            "test/requirements/with-requirements-txt/constraints_out.yaml",
            ["merlin-pyfunc-server<0.42.0"],
        ),
        # Test using dictionary
        # Empty conda env
        (
            {},
            {
                "dependencies": [
                    "python=3.10.*",
                    {"pip": ["merlin-pyfunc-server<0.42.0"]},
                ]
            },
            ["merlin-pyfunc-server<0.42.0"],
        ),
        # Using old python version, no pip reqs
        (
            {
                "dependencies": [
                    "python=3.7.*",
                ]
            },
            {
                "dependencies": [
                    "python=3.10.*",
                    {"pip": ["merlin-pyfunc-server<0.42.0"]},
                ]
            },
            ["merlin-pyfunc-server<0.42.0"],
        ),
        # Using old python version, with empty pip reqs
        (
            {
                "dependencies": [
                    "python=3.7.*",
                    {"pip": []},
                ]
            },
            {
                "dependencies": [
                    "python=3.10.*",
                    {"pip": ["merlin-pyfunc-server<0.42.0"]},
                ]
            },
            ["merlin-pyfunc-server<0.42.0"],
        ),
        # pip reqs has merlin-pyfunc-server without version
        (
            {
                "dependencies": [
                    "python=3.10.*",
                    {"pip": ["merlin-pyfunc-server"]},
                ]
            },
            {
                "dependencies": [
                    "python=3.10.*",
                    {"pip": ["merlin-pyfunc-server"]},
                ]
            },
            ["merlin-pyfunc-server<0.42.0"],
        ),
        # pip reqs has merlin-pyfunc-server with older version, should not be updated
        (
            {
                "dependencies": [
                    "python=3.10.*",
                    {"pip": ["merlin-pyfunc-server==0.20.0"]},
                ]
            },
            {
                "dependencies": [
                    "python=3.10.*",
                    {"pip": ["merlin-pyfunc-server==0.20.0"]},
                ]
            },
            ["merlin-pyfunc-server<0.42.0"],
        ),
        # pip reqs has merlin-pyfunc-server with newer version, should not be updated
        (
            {
                "dependencies": [
                    "python=3.10.*",
                    {"pip": ["merlin-pyfunc-server==0.50.0"]},
                ]
            },
            {
                "dependencies": [
                    "python=3.10.*",
                    {"pip": ["merlin-pyfunc-server==0.50.0"]},
                ]
            },
            ["merlin-pyfunc-server<0.42.0"],
        ),
    ],
)
def test_process_conda_env(input, output, additional_merlin_reqs):
    default_python_version = "3.10.*"

    actual_conda_env = process_conda_env(
        conda_env=input,
        python_version=default_python_version,
        additional_merlin_reqs=additional_merlin_reqs,
    )

    expected_conda_env = output
    if isinstance(output, str):
        with open(output, "r") as f:
            expected_conda_env = yaml.safe_load(f)

    assert actual_conda_env == expected_conda_env



import torch.nn as nn
import torch.nn.functional as F

class IrisModel(nn.Module):
    # define nn
    def __init__(self):
        super(IrisModel, self).__init__()
        self.fc1 = nn.Linear(4, 100)
        self.fc2 = nn.Linear(100, 100)
        self.fc3 = nn.Linear(100, 3)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, X):
        X = F.relu(self.fc1(X))
        X = self.fc2(X)
        X = self.fc3(X)
        X = self.softmax(X)

        return X


# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from contextlib import contextmanager
from typing import Any, Dict, List, Optional

from client import PredictionJob
from merlin.autoscaling import AutoscalingPolicy
from merlin.batch.config import PredictionJobConfig
from merlin.client import MerlinClient
from merlin.deployment_mode import DeploymentMode
from merlin.endpoint import ModelEndpoint, VersionEndpoint
from merlin.environment import Environment
from merlin.logger import Logger
from merlin.model import Model, ModelType, ModelVersion, Project
from merlin.model_schema import ModelSchema
from merlin.protocol import Protocol
from merlin.resource_request import ResourceRequest
from merlin.transformer import Transformer
from merlin.version_image import VersionImage
from merlin.model_observability import ModelObservability

_merlin_client: Optional[MerlinClient] = None
_active_project: Optional[Project]
_active_model: Optional[Model]
_active_model_version: Optional[ModelVersion]


def set_url(url: str, use_google_oauth: bool = True):
    """
    Set Merlin URL

    :param url: Merlin URL
    """
    global _merlin_client
    _merlin_client = MerlinClient(url, use_google_oauth)


def get_url() -> Optional[str]:
    """
    Get currently active Merlin URL

    :return: merlin url if set, otherwise None
    """
    if _merlin_client is not None:
        return _merlin_client.url
    return None


def list_environment() -> List[Environment]:
    """
    List all available environment for deployment

    :return: List[Environment]
    """
    _check_active_client()
    return _merlin_client.list_environment()  # type: ignore


def get_environment(env_name: str) -> Environment:
    """
    Get environment for given env name

    :return: Environment or None
    """
    _check_active_client()
    envs = _merlin_client.list_environment()  # type: ignore
    for env in envs:
        if env.name == env_name:
            return env
    return None  # type: ignore


def get_default_environment() -> Optional[Environment]:
    """
    Get default environment

    :return: Environment or None
    """
    _check_active_client()
    return _merlin_client.get_default_environment()  # type: ignore


def list_project() -> List[Project]:
    """
    List all project in MLP

    :return: list of project
    """
    _check_active_client()
    return _merlin_client.list_project()  # type: ignore


def set_project(project_name: str):
    """
    Set active project

    :param project_name: project name. If project_name is not found, it will create the project.
    """
    _check_active_client()

    p = _merlin_client.get_project(project_name)  # type: ignore
    global _active_project
    _active_project = p


def active_project() -> Optional[Project]:
    """
    Get current active project

    :return: active project
    """
    _check_active_client()
    _check_active_project()
    return _active_project


def set_model(model_name, model_type: ModelType = None):
    """
    Set active model

    :param model_name: model name to be set as active model. If model name is not found, it will create the model.
    :param model_type: type of the model
    :return:
    """
    _check_active_project()
    active_project_name = _active_project.name  # type: ignore
    mdl = _merlin_client.get_or_create_model(  # type: ignore
        model_name, active_project_name, model_type
    )
    global _active_model
    _active_model = mdl


def active_model() -> Optional[Model]:
    """
    Get active model

    :return: active model
    """
    _check_active_client()
    _check_active_project()
    _check_active_model()

    return _active_model


@contextmanager
def new_model_version(
    labels: Dict[str, str] = None, model_schema: Optional[ModelSchema] = None
):
    """
    Create new model version under currently active model

    :param labels: dictionary containing the label that will be stored in the new model version
    :param model_schema: Detail schema specification of a model
    :return: ModelVersion
    """
    v = None
    try:
        _check_active_client()
        _check_active_project()
        _check_active_model()
        v = _merlin_client.new_model_version(_active_model.name, _active_project.name, labels, model_schema)  # type: ignore
        v.start()
        global _active_model_version
        _active_model_version = v
        yield v
    finally:
        _active_model_version = None
        if v is not None:
            v.finish()


def log_param(key: str, value: str):
    """
    Log parameter to the active model version

    :param key: parameter key
    :param value: parameter value
    """
    _check_active_model_version()
    _active_model_version.log_param(key, value)  # type: ignore


def log_metric(key: str, value: float):
    """
    Log a metric to the active model version

    :param key: metric key
    :param value: metric value
    """
    _check_active_model_version()
    _active_model_version.log_metric(key, value)  # type: ignore


def set_tag(key: str, value: str):
    """
    Set tag in the active model version

    :param key: tag name
    :param value: tag value
    """
    _check_active_model_version()
    _active_model_version.set_tag(key, value)  # type: ignore


def delete_tag(key: str):
    """
    Delete tag from the active model version

    :param key: tag name
    """
    _check_active_model_version()
    _active_model_version.delete_tag(key)  # type: ignore


def get_param(key: str) -> Optional[str]:
    """
    Get param value from the active model version

    :param key: param name
    """
    _check_active_model_version()
    return _active_model_version.get_param(key)  # type: ignore


def get_metric(key: str) -> Optional[float]:
    """
    Get metric value from the active model version

    :param key: metric name
    """
    _check_active_model_version()
    return _active_model_version.get_metric(key)  # type: ignore


def get_tag(key: str) -> Optional[str]:
    """
    Get tag value from the active model version

    :param key: tag name
    """
    _check_active_model_version()
    return _active_model_version.get_tag(key)  # type: ignore


def list_tag() -> Dict[str, str]:
    """
    Get list of tag from the active model version
    """
    _check_active_model_version()
    return _active_model_version.list_tag()  # type: ignore


def download_artifact(destination_path: str):
    """
    Download artifact from the active model version

    :param destination_path: destination of file when downloaded
    """
    _check_active_model_version()
    _active_model_version.download_artifact(destination_path)  # type: ignore


def log_artifact(local_path: str, artifact_path: str = None):
    """
    Log artifacts for the active model version

    :param local_path: directory to be uploaded into artifact store
    :param artifact_path: destination directory in artifact store
    """
    _check_active_model_version()
    _active_model_version.log_artifact(local_path, artifact_path)  # type: ignore


def log_pyfunc_model(
    model_instance: Any,
    conda_env: str,
    code_dir: List[str] = None,
    artifacts: Dict[str, str] = None,
):
    """
    Upload PyFunc based model into artifact storage.

    User has to specify model_instance and
    conda_env. `model_instance` shall implement all method specified in
    PyFuncModel. conda_env shall contain all dependency required by the
    model


    :param model_instance: instance of python function model
    :param conda_env: path to conda env.yaml file
    :param code_dir: additional code directory that will be loaded with ModelType.PYFUNC model
    :param artifacts: dictionary of artifact that will be stored together with the model. This will be passed to PythonModel.initialize. Example: {"config": "config/staging.yaml"}
    """
    _check_active_model_version()
    _active_model_version.log_pyfunc_model(  # type: ignore
        model_instance, conda_env, code_dir, artifacts
    )


def log_pytorch_model(model_dir: str, model_class_name: str = None):
    """
    Upload PyTorch model to artifact storage.

    :param model_dir: directory containing serialized PyTorch model
    :param model_class_name: class name of PyTorch model. By default the model class name is 'PyTorchModel'
    """
    _check_active_model_version()
    _active_model_version.log_pytorch_model(model_dir, model_class_name)  # type: ignore


def log_model(model_dir):
    """
    Upload model to artifact storage.
    This method is used to upload model for xgboost, tensorflow, and sklearn model.

    :param model_dir: directory which contain serialized model
    """
    _check_active_model_version()
    _active_model_version.log_model(model_dir)  # type: ignore


def log_custom_model(
    image: str, model_dir: str = None, command: str = "", args: str = ""
):
    """
    Upload model to artifact storage.
    This method is used to upload model for custom model type.

    :param image: Docker image that will be used as predictor
    :param model_dir: directory which contain serialized model
    :param command: Command to run docker image
    :param args: Arguments that needs to be specified when running docker
    """
    _check_active_model_version()
    _active_model_version.log_custom_model(  # type: ignore
        image=image, model_dir=model_dir, command=command, args=args
    )


def build_image(
    model_version: ModelVersion = None,
    backoff_limit: int = 0,
    resource_request: ResourceRequest = None,
) -> VersionImage:
    """
    Build image for a model version

    :param model_version: If model_version is not given it will build image for active model version
    :param backoff_limit: The maximum number of retries before considering a job as failed.
    :param resource_request: The resource requirement (CPU & memory) for image builder job.
    :return: VersionImage
    """
    _check_active_client()
    if model_version is None:
        _check_active_model_version()
        return _active_model_version.build_image(backoff_limit=backoff_limit, resource_request=resource_request)  # type: ignore

    return _merlin_client.build_image(model_version=model_version, backoff_limit=backoff_limit, resource_request=resource_request)  # type: ignore


def deploy(
    model_version: ModelVersion = None,
    environment_name: str = None,
    resource_request: ResourceRequest = None,
    image_builder_resource_request: ResourceRequest = None,
    env_vars: Dict[str, str] = None,
    transformer: Transformer = None,
    logger: Logger = None,
    deployment_mode: DeploymentMode = None,
    autoscaling_policy: AutoscalingPolicy = None,
    protocol: Protocol = None,
    enable_model_observability: bool = False,
    model_observability: Optional[ModelObservability] = None,
) -> VersionEndpoint:
    """
    Deploy a model version.

    :param model_version: If model_version is not given it will deploy active model version
    :param environment_name: target environment to which the model version will be deployed to. If left empty it will deploy to default environment.
    :param resource_request: The resource requirement and replicas requests for model version endpoint.
    :param image_builder_resource_request: The resource requirement and replicas requests for image builder job.
    :param env_vars: List of environment variables to be passed to the model container.
    :param transformer: The service to be deployed alongside the model for pre/post-processing steps.
    :param logger: Response/Request logging configuration for model or transformer.
    :param deployment_mode: mode of deployment for the endpoint (default: DeploymentMode.SERVERLESS)
    :param autoscaling_policy: autoscaling policy to be used for the deployment (default: None)
    :param protocol: protocol to be used by the deployed model (default: HTTP_JSON)
    :param enable_model_observability: flag to determine whether model observability enabled for the endpoint
    :param model_observability: detail of model observability configuration
    :return: VersionEndpoint object
    """
    _check_active_client()
    if model_version is None:
        _check_active_model_version()
        return _active_model_version.deploy(  # type: ignore
            environment_name,
            resource_request,
            image_builder_resource_request,
            env_vars,
            transformer,
            logger,
            deployment_mode,
            autoscaling_policy,
            protocol,
            enable_model_observability,
            model_observability
        )

    return _merlin_client.deploy(  # type: ignore
        model_version,
        environment_name,
        resource_request,
        image_builder_resource_request,
        env_vars,
        transformer,
        logger,
        deployment_mode,
        autoscaling_policy,
        protocol,
        enable_model_observability,
        model_observability
    )


def undeploy(model_version=None, environment_name: str = None):
    """
    Delete deployment of a model version.

    :param model_version: model version to be undeployed. If model_version is not given it will undeploy active model version
    """
    _check_active_client()
    if model_version is None:
        _check_active_model_version()
        _active_model_version.undeploy(environment_name)  # type: ignore
        return

    _merlin_client.undeploy(model_version, environment_name)  # type: ignore


def serve_traffic(
    traffic_rule: Dict["VersionEndpoint", int], environment_name: str = None
) -> ModelEndpoint:
    """
    Update traffic rule of the active model.

    :param traffic_rule: dict of version endpoint and the percentage of traffic.
    :param environment_name: environment in which the traffic rule shall be applied
    :return: ModelEndpoint
    """
    _check_active_model()
    return _active_model.serve_traffic(traffic_rule, environment_name)  # type: ignore


def stop_serving_traffic(environment_name: str = None):
    """
    Stop serving traffic for a given model endpoint in given environment.

    :param environment_name: environment in which the model endpoint will be stopped.
    """
    _check_active_model()
    return _active_model.stop_serving_traffic(environment_name)  # type: ignore


def set_traffic(traffic_rule: Dict[ModelVersion, int]) -> ModelEndpoint:
    """
     Update traffic rule of the active model.

    :param traffic_rule: dict of model version and the percentage of traffic.
    :return: ModelEndpoint
    """
    _check_active_model()
    return _active_model.set_traffic(traffic_rule)  # type: ignore


def list_model_endpoints() -> List[ModelEndpoint]:
    """
    Get list of all serving model endpoints.

    :return: List of model endpoints.
    """
    _check_active_model()
    return _active_model.list_endpoint()  # type: ignore


def create_prediction_job(
    job_config: PredictionJobConfig, sync: bool = True
) -> PredictionJob:
    """

    :param sync:
    :param job_config:
    :return:
    """
    _check_active_client()
    _check_active_project()
    _check_active_model()
    _check_active_model_version()

    return _active_model_version.create_prediction_job(job_config=job_config, sync=sync)  # type: ignore


def _check_active_project():
    if _active_project is None:
        raise Exception("Active project isn't set, use set_project(...) to set it")


def _check_active_client():
    if _merlin_client is None:
        raise Exception("URL is not set, use set_url(...) to set it")


def _check_active_model():
    if _active_model is None:
        raise Exception("Active model isn't set, use set_model(...) to set it")


def _check_active_model_version():
    if _active_model_version is None:
        raise Exception(
            "Active model version isn't set, use new_model_version(...) to " "create it"
        )



import os
from collections import namedtuple
from itertools import filterfalse
from pathlib import Path
from typing import Any, List, Optional

import yaml
from packaging.requirements import Requirement, InvalidRequirement

_CONSTRAINTS_FILE_NAME = "constraints.txt"


def process_conda_env(
    conda_env: Any,
    python_version: Optional[str] = "3.10.*",
    additional_merlin_reqs: Optional[List[str]] = None,
):
    """
    Processes `conda_env` passed to `merlin.log_pyfunc_model` and `merlin.log_model`.
    This function will do the following:
    1. Add/replace Python version on the conda_env with `python_version` argument.
    2. Add `additional_merlin_reqs` onto the conda_env's pip dependencies.

    :param conda_env: Either a dictionary representation of a conda environment or the path to a conda environment yaml file
    :param python_version: The python version to replace the conda environment with
    :param additonal_merlin_reqs: A list of additional Merlin requirements to be added to the conda environment
    :return: dict representation of the conda environment
    """

    if isinstance(conda_env, str):
        with open(conda_env, "r") as f:
            conda_env = yaml.safe_load(f)
    elif not isinstance(conda_env, dict):
        raise TypeError(
            "Expected a string path to a conda env yaml file or a `dict` representing a conda env, "
            "but got `{}`".format(type(conda_env).__name__)
        )

    if conda_env is None:
        conda_env = {}

    if "dependencies" not in conda_env:
        conda_env["dependencies"] = []

    conda_env = _overwrite_python_version(conda_env, python_version)

    pip_reqs = _get_pip_deps(conda_env)
    pip_reqs, constraints = _parse_pip_requirements(pip_reqs)

    if additional_merlin_reqs:
        for additional_merlin_req in additional_merlin_reqs:
            exist = False
            for pip_req in pip_reqs:
                # Skip requirement comparison for pip reqs that is not a package, which can happen if the user specifies
                # repository or trusted host as part of the pip file
                if pip_req.strip().startswith("--"):
                    continue

                pip_req_obj = Requirement(pip_req)
                additional_merlin_req_obj = Requirement(additional_merlin_req)
                if pip_req_obj.name.lower() == additional_merlin_req_obj.name.lower():
                    exist = True
                    break

            if not exist:
                pip_reqs.insert(0, additional_merlin_req)

    if constraints:
        pip_reqs.append(f"-c {_CONSTRAINTS_FILE_NAME}")

    conda_env = _overwrite_pip_deps(conda_env, pip_reqs)

    return conda_env


def _overwrite_python_version(conda_env, python_version):
    """
    Overwrites the python version in the given conda env dictionary.

    {
        "name": "env",
        "channels": [...],
        "dependencies": [
            ...,
            "python=3.10.*",  <- Overwrite this
        ],
    }
    """
    for dep in conda_env["dependencies"]:
        if isinstance(dep, str) and "python" in dep:
            conda_env["dependencies"].remove(dep)

    conda_env["dependencies"].insert(0, "python={}".format(python_version))
    return conda_env


def _get_pip_deps(conda_env):
    """
    :return: The pip dependencies from the conda env
    """
    if conda_env is not None:
        for dep in conda_env["dependencies"]:
            if _is_pip_deps(dep):
                return dep["pip"]
    return []


def _parse_pip_requirements(pip_requirements):
    """
    Parses an iterable of pip requirement strings or a pip requirements file.

    :param pip_requirements: Either an iterable of pip requirement strings
        (e.g. ``["scikit-learn", "-r requirements.txt"]``) or the string path to a pip requirements
        file on the local filesystem (e.g. ``"requirements.txt"``). If ``None``, an empty list will
        be returned.
    :return: A tuple of parsed requirements and constraints.
    """
    if pip_requirements is None:
        return [], []

    def _is_string(x):
        return isinstance(x, str)

    def _is_iterable(x):
        try:
            iter(x)
            return True
        except Exception:
            return False

    if _is_string(pip_requirements):
        with open(pip_requirements) as f:
            return _parse_pip_requirements(f.read().splitlines())
    elif _is_iterable(pip_requirements) and all(map(_is_string, pip_requirements)):
        requirements = []
        constraints = []
        for req_or_con in _parse_requirements(pip_requirements, is_constraint=False):
            if req_or_con.is_constraint:
                constraints.append(req_or_con.req_str)
            else:
                requirements.append(req_or_con.req_str)

        return requirements, constraints
    else:
        raise TypeError(
            "`pip_requirements` must be either a string path to a pip requirements file on the "
            "local filesystem or an iterable of pip requirement strings, but got `{}`".format(
                type(pip_requirements)
            )
        )


# Represents a pip requirement.
#
# :param req_str: A requirement string (e.g. "scikit-learn == 0.24.2").
# :param is_constraint: A boolean indicating whether this requirement is a constraint.
_Requirement = namedtuple("_Requirement", ["req_str", "is_constraint"])


def _parse_requirements(requirements, is_constraint, base_dir=None):
    """
    A simplified version of `pip._internal.req.parse_requirements` which performs the following
    operations on the given requirements file and yields the parsed requirements.

    - Remove comments and blank lines
    - Join continued lines
    - Resolve requirements file references (e.g. '-r requirements.txt')
    - Resolve constraints file references (e.g. '-c constraints.txt')

    :param requirements: A string path to a requirements file on the local filesystem or
                         an iterable of pip requirement strings.
    :param is_constraint: Indicates the parsed requirements file is a constraint file.
    :param base_dir: If specified, resolve relave file references (e.g. '-r requirements.txt')
                     against the specified directory.
    :return: A list of ``_Requirement`` instances.

    References:
    - `pip._internal.req.parse_requirements`:
      https://github.com/pypa/pip/blob/7a77484a492c8f1e1f5ef24eaf71a43df9ea47eb/src/pip/_internal/req/req_file.py#L118
    - Requirements File Format:
      https://pip.pypa.io/en/stable/cli/pip_install/#requirements-file-format
    - Constraints Files:
      https://pip.pypa.io/en/stable/user_guide/#constraints-files
    """
    if base_dir is None:
        if isinstance(requirements, (str, Path)):
            base_dir = os.path.dirname(requirements)
            with open(requirements) as f:
                requirements = f.read().splitlines()
        else:
            base_dir = os.getcwd()

    lines = map(str.strip, requirements)
    lines = map(_strip_inline_comment, lines)
    lines = _join_continued_lines(lines)
    lines = filterfalse(_is_comment, lines)
    lines = filterfalse(_is_empty, lines)

    for line in lines:
        if _is_requirements_file(line):
            req_file = line.split(maxsplit=1)[1]
            # If `req_file` is an absolute path, `os.path.join` returns `req_file`:
            # https://docs.python.org/3/library/os.path.html#os.path.join
            abs_path = os.path.join(base_dir, req_file)
            yield from _parse_requirements(abs_path, is_constraint=False)
        elif _is_constraints_file(line):
            req_file = line.split(maxsplit=1)[1]
            abs_path = os.path.join(base_dir, req_file)
            yield from _parse_requirements(abs_path, is_constraint=True)
        else:
            yield _Requirement(line, is_constraint)


def _is_comment(line):
    return line.startswith("#")


def _is_empty(line):
    return line == ""


def _strip_inline_comment(line):
    return line[: line.find(" #")].rstrip() if " #" in line else line


def _is_requirements_file(line):
    return line.startswith("-r ") or line.startswith("--requirement ")


def _is_constraints_file(line):
    return line.startswith("-c ") or line.startswith("--constraint ")


def _join_continued_lines(lines):
    """
    Joins lines ending with '\\'.

    >>> _join_continued_lines["a\\", "b\\", "c"]
    >>> 'abc'
    """
    continued_lines = []

    for line in lines:
        if line.endswith("\\"):
            continued_lines.append(line.rstrip("\\"))
        else:
            continued_lines.append(line)
            yield "".join(continued_lines)
            continued_lines.clear()

    # The last line ends with '\'
    if continued_lines:
        yield "".join(continued_lines)


def _is_pip_deps(dep):
    """
    Returns True if `dep` is a dict representing pip dependencies
    """
    return isinstance(dep, dict) and "pip" in dep


def _overwrite_pip_deps(conda_env, new_pip_deps):
    """
    Overwrites the pip dependencies section in the given conda env dictionary.

    {
        "name": "env",
        "channels": [...],
        "dependencies": [
            ...,
            "pip",
            {"pip": [...]},  <- Overwrite this
        ],
    }
    """
    deps = conda_env.get("dependencies", [])
    new_deps = []
    contains_pip_deps = False
    for dep in deps:
        if _is_pip_deps(dep):
            contains_pip_deps = True
            new_deps.append({"pip": new_pip_deps})
        else:
            new_deps.append(dep)

    if not contains_pip_deps:
        new_deps.append({"pip": new_pip_deps})

    return {**conda_env, "dependencies": new_deps}



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from sys import version_info
from typing import Any, Dict, List, Optional

import client
import urllib3
from caraml_auth.id_token_credentials import get_default_id_token_credentials
from client import (
    ApiClient,
    Configuration,
    EndpointApi,
    EnvironmentApi,
    ModelsApi,
    ProjectApi,
    StandardTransformerApi,
    StandardTransformerSimulationRequest,
    VersionApi,
)
from google.auth.transport.requests import Request
from google.auth.transport.urllib3 import AuthorizedHttp
from merlin.autoscaling import AutoscalingPolicy
from merlin.deployment_mode import DeploymentMode
from merlin.endpoint import VersionEndpoint
from merlin.environment import Environment
from merlin.logger import Logger
from merlin.model import Model, ModelType, ModelVersion, Project
from merlin.model_schema import ModelSchema
from merlin.protocol import Protocol
from merlin.resource_request import ResourceRequest
from merlin.transformer import Transformer
from merlin.util import valid_name_check
from merlin.version import VERSION
from merlin.version_image import VersionImage
from merlin.model_observability import ModelObservability


class MerlinClient:
    def __init__(self, merlin_url: str, use_google_oauth: bool = True):
        self._merlin_url = merlin_url
        config = Configuration()
        config.host = self._merlin_url + "/v1"

        self._api_client = ApiClient(config)
        if use_google_oauth:
            credentials = get_default_id_token_credentials(target_audience="sdk.caraml")
            # Refresh credentials, in case it's coming from Compute Engine.
            # See: https://github.com/googleapis/google-auth-library-python/issues/1211
            credentials.refresh(Request())
            authorized_http = AuthorizedHttp(credentials, urllib3.PoolManager())
            self._api_client.rest_client.pool_manager = authorized_http  # type: ignore

        python_version = f"{version_info.major}.{version_info.minor}.{version_info.micro}"  # capture user's python version
        self._api_client.user_agent = f"merlin-sdk/{VERSION} python/{python_version}"
        self._project_api = ProjectApi(self._api_client)
        self._model_api = ModelsApi(self._api_client)
        self._version_api = VersionApi(self._api_client)
        self._endpoint_api = EndpointApi(self._api_client)
        self._env_api = EnvironmentApi(self._api_client)
        self._standard_transformer_api = StandardTransformerApi(self._api_client)

    @property
    def url(self):
        return self._merlin_url

    def list_environment(self) -> List[Environment]:
        """
        List all available environment for deployment

        :return: list of Environment
        """
        env_list = self._env_api.environments_get()
        envs = []
        for env in env_list:
            envs.append(Environment(env))
        return envs

    def get_environment(self, env_name: str) -> Optional[Environment]:
        """
        Get environment for given env name

        :return: Environment or None
        """
        env_list = self._env_api.environments_get()
        for env in env_list:
            if env.name == env_name:
                return Environment(env)
        return None

    def get_default_environment(self) -> Optional[Environment]:
        """
        Return default environment

        :return: Environment or None
        """
        env_list = self._env_api.environments_get()
        for env in env_list:
            if env.is_default:
                return Environment(env)
        return None

    def list_project(self) -> List[Project]:
        """
        List project in the connected MLP server

        :return: list of Project
        """
        p_list = self._project_api.projects_get()
        result = []
        for p in p_list:
            result.append(Project(p, self.url, self._api_client))
        return result

    def get_or_create_project(self, project_name: str) -> Project:
        warnings.warn(
            "get_or_create_project is deprecated please use get_project",
            category=DeprecationWarning,
        )
        return self.get_project(project_name)

    def get_project(self, project_name: str) -> Project:
        """
        Get a project in Merlin and optionally assign list of readers and administrators.
        The identity used for creating the project will be automatically included as project's administrators.

        :param project_name: project name
        :return: project
        """
        if not valid_name_check(project_name):
            raise ValueError(
                """Your project/model name contains invalid characters.\
                    \nUse only the following characters\
                    \n- Characters: a-z (Lowercase ONLY)\
                    \n- Numbers: 0-9\
                    \n- Symbols: -
                """
            )

        p_list = self._project_api.projects_get(name=project_name)
        p = None
        for prj in p_list:
            if prj.name == project_name:
                p = prj

        if p is None:
            raise Exception(
                f"{project_name} does not exist or you don't have access to the project. Please create new "
                f"project using MLP console or ask the project's administrator to be able to access "
                f"existing project."
            )

        return Project(p, self.url, self._api_client)

    def get_model(self, model_name: str, project_name: str) -> Optional[Model]:
        """
        Get model with given name

        :param model_name: model name to be retrieved
        :param project_name: project name
        :return: Model or None
        """
        prj = self.get_project(project_name)
        m_list = self._model_api.projects_project_id_models_get(
            project_id=int(prj.id), name=model_name
        )
        model = m_list[0]
        if len(m_list) == 0:
            return None

        for mdl in m_list:
            if mdl.name == model_name:
                model = mdl

        return Model(model, prj, self._api_client)

    def get_or_create_model(
        self, model_name: str, project_name: str, model_type: ModelType = None
    ) -> Model:
        """
        Get or create a model under a project

        If project_name is not given it will use currently active project
        otherwise will raise Exception

        :param model_type:
        :param model_name: model name
        :param project_name: project name (optional)
        :param model_type: model type
        :return: Model
        """
        if not valid_name_check(model_name):
            raise ValueError(
                """Your project/model name contains invalid characters.\
                    \nUse only the following characters\
                    \n- Characters: a-z (Lowercase ONLY)\
                    \n- Numbers: 0-9\
                    \n- Symbols: -
                """
            )

        prj = self.get_project(project_name)
        m_list = self._model_api.projects_project_id_models_get(
            project_id=int(prj.id), name=model_name
        )

        model = None
        for mdl in m_list:
            if mdl.name == model_name:
                model = mdl

        if model is None:
            if model_type is None:
                raise ValueError(
                    f"model {model_name} is not found, specify "
                    f"{model_type} to create it"
                )
            model = self._model_api.projects_project_id_models_post(
                project_id=int(prj.id),
                body=client.Model(name=model_name, type=model_type.value),
            )

        return Model(model, prj, self._api_client)

    def new_model_version(
        self,
        model_name: str,
        project_name: str,
        labels: Dict[str, str] = None,
        model_schema: Optional[ModelSchema] = None,
    ) -> ModelVersion:
        """
        Create new model version for the given model and project

        :param model_name:
        :param project_name:
        :param labels:
        :return:
        """
        mdl = self.get_model(model_name, project_name)
        if mdl is None:
            raise ValueError(f"Model with name: {model_name} is not found")
        return mdl.new_model_version(labels=labels, model_schema=model_schema)

    def build_image(
        self,
        model_version: ModelVersion,
        backoff_limit: int = 0,
        resource_request: ResourceRequest = None,
    ) -> VersionImage:
        return model_version.build_image(
            backoff_limit=backoff_limit,
            resource_request=resource_request,
        )

    def deploy(
        self,
        model_version: ModelVersion,
        environment_name: str = None,
        resource_request: ResourceRequest = None,
        image_builder_resource_request: ResourceRequest = None,
        env_vars: Dict[str, str] = None,
        transformer: Transformer = None,
        logger: Logger = None,
        deployment_mode: DeploymentMode = None,
        autoscaling_policy: AutoscalingPolicy = None,
        protocol: Protocol = None,
        enable_model_observability: bool = False,
        model_observability: Optional[ModelObservability] = None
    ) -> VersionEndpoint:
        return model_version.deploy(
            environment_name,
            resource_request,
            image_builder_resource_request,
            env_vars,
            transformer,
            logger,
            deployment_mode,
            autoscaling_policy,
            protocol,
            enable_model_observability,
            model_observability
        )

    def undeploy(self, model_version: ModelVersion, environment_name: str = None):
        model_version.undeploy(environment_name)

    def standard_transformer_simulate(
        self,
        payload: Dict,
        headers: Optional[Dict[Any, Any]] = None,
        config: Optional[Dict[Any, Any]] = None,
        model_prediction_config: Dict = None,
        protocol: str = "HTTP_JSON",
    ):
        prediction_config = (
            client.ModelPredictionConfig.from_dict(model_prediction_config)
            if model_prediction_config is not None
            else None
        )
        request = StandardTransformerSimulationRequest(
            payload=payload,
            headers=headers,
            config=config,
            model_prediction_config=prediction_config,
            protocol=client.Protocol(protocol),
        )

        return self._standard_transformer_api.standard_transformer_simulate_post(
            body=request
        )



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Dict, Optional

from merlin.batch.sink import Sink
from merlin.batch.source import Source
from merlin.resource_request import ResourceRequest


class ResultType(Enum):
    DOUBLE = "DOUBLE"
    FLOAT = "FLOAT"
    INTEGER = "INTEGER"
    LONG = "LONG"
    STRING = "STRING"
    ARRAY = "ARRAY"


class PredictionJobResourceRequest:
    """
    Resource request configuration for starting prediction job
    """

    def __init__(
        self,
        driver_cpu_request: str,
        driver_memory_request: str,
        executor_cpu_request: str,
        executor_memory_request: str,
        executor_replica: int,
    ):
        """
        Create resource request object

        :param driver_cpu_request: driver's cpu request in kubernetes request format (e.g. : 500m, 1, 2, etc)
        :param driver_memory_request: driver's memory request in kubernetes format (e.g.: 512Mi, 1Gi, 2Gi, etc)
        :param executor_cpu_request: executors's cpu request in kubernetes request format (e.g. : 500m, 1, 2, etc)
        :param executor_memory_request: executors's memory request in kubernetes format (e.g.: 512Mi, 1Gi, 2Gi, etc)
        :param executor_replica: number of executor to be used
        """
        self._driver_cpu_request = driver_cpu_request
        self._driver_memory_request = driver_memory_request
        self._executor_cpu_request = executor_cpu_request
        self._executor_memory_request = executor_memory_request
        self._executor_replica = executor_replica

    def to_dict(self):
        return {
            "driver_cpu_request": self._driver_cpu_request,
            "driver_memory_request": self._driver_memory_request,
            "executor_cpu_request": self._executor_cpu_request,
            "executor_memory_request": self._executor_memory_request,
            "executor_replica": self._executor_replica,
        }


class PredictionJobConfig:
    def __init__(
        self,
        source: Source,
        sink: Sink,
        service_account_name: str,
        result_type: ResultType = ResultType.DOUBLE,
        item_type: ResultType = ResultType.DOUBLE,
        resource_request: PredictionJobResourceRequest = None,
        image_builder_resource_request: ResourceRequest = None,
        env_vars: Dict[str, str] = None,
    ):
        """
        Create configuration for starting a prediction job

        :param source: source configuration. See merlin.batch.source package.
        :param sink: sink configuration. See merlin.batch.sink package
        :param service_account_name: secret name containing the service account for executing the prediction job.
        :param result_type: type of the prediction result (default to ResultType.DOUBLE).
        :param item_type: item type of the prediction result if the result_type is ResultType.ARRAY. Otherwise will be ignored.
        :param resource_request: optional resource request for starting the prediction job. If not given the system default will be used.
        :param image_builder_resource_request: optional resource request for image builder job.
        :param env_vars: optional environment variables in the form of a key value pair in a list.
        """

        self._source = source
        self._sink = sink
        self._service_account_name = service_account_name
        self._resource_request = resource_request
        self._image_builder_resource_request = image_builder_resource_request
        self._result_type = result_type
        self._item_type = item_type
        self._env_vars = env_vars

    @property
    def source(self) -> Source:
        return self._source

    @property
    def sink(self) -> Sink:
        return self._sink

    @property
    def service_account_name(self) -> str:
        return self._service_account_name

    @property
    def resource_request(self) -> Optional[PredictionJobResourceRequest]:
        return self._resource_request

    @property
    def image_builder_resource_request(self) -> Optional[ResourceRequest]:
        return self._image_builder_resource_request

    @property
    def result_type(self) -> ResultType:
        return self._result_type

    @property
    def item_type(self) -> ResultType:
        return self._item_type

    @property
    def env_vars(self) -> Optional[Dict[str, str]]:
        return self._env_vars



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictInt, StrictStr, field_validator
from pydantic import Field
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class Model(BaseModel):
    """
    Model
    """ # noqa: E501
    id: Optional[StrictInt] = None
    project_id: Optional[StrictInt] = None
    mlflow_experiment_id: Optional[StrictInt] = None
    name: StrictStr
    type: Optional[StrictStr] = Field(default=None, description="Model type")
    mlflow_url: Optional[StrictStr] = None
    endpoints: Optional[List[ModelEndpoint]] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    __properties: ClassVar[List[str]] = ["id", "project_id", "mlflow_experiment_id", "name", "type", "mlflow_url", "endpoints", "created_at", "updated_at"]

    @field_validator('type')
    def type_validate_enum(cls, value):
        """Validates the enum"""
        if value is None:
            return value

        if value not in ('xgboost', 'tensorflow', 'sklearn', 'pytorch', 'pyfunc', 'pyfunc_v2', 'custom'):
            raise ValueError("must be one of enum values ('xgboost', 'tensorflow', 'sklearn', 'pytorch', 'pyfunc', 'pyfunc_v2', 'custom')")
        return value

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of Model from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in endpoints (list)
        _items = []
        if self.endpoints:
            for _item in self.endpoints:
                if _item:
                    _items.append(_item.to_dict())
            _dict['endpoints'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of Model from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "id": obj.get("id"),
            "project_id": obj.get("project_id"),
            "mlflow_experiment_id": obj.get("mlflow_experiment_id"),
            "name": obj.get("name"),
            "type": obj.get("type"),
            "mlflow_url": obj.get("mlflow_url"),
            "endpoints": [ModelEndpoint.from_dict(_item) for _item in obj.get("endpoints")] if obj.get("endpoints") is not None else None,
            "created_at": obj.get("created_at"),
            "updated_at": obj.get("updated_at")
        })
        return _obj

from client.models.model_endpoint import ModelEndpoint
# TODO: Rewrite to not use raise_errors
Model.model_rebuild(raise_errors=False)




# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictBool, StrictInt, StrictStr
from client.models.autoscaling_policy import AutoscalingPolicy
from client.models.deployment_mode import DeploymentMode
from client.models.endpoint_status import EndpointStatus
from client.models.env_var import EnvVar
from client.models.environment import Environment
from client.models.logger import Logger
from client.models.model_observability import ModelObservability
from client.models.protocol import Protocol
from client.models.resource_request import ResourceRequest
from client.models.transformer import Transformer
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class VersionEndpoint(BaseModel):
    """
    VersionEndpoint
    """ # noqa: E501
    id: Optional[StrictStr] = None
    version_id: Optional[StrictInt] = None
    status: Optional[EndpointStatus] = None
    url: Optional[StrictStr] = None
    service_name: Optional[StrictStr] = None
    environment_name: Optional[StrictStr] = None
    environment: Optional[Environment] = None
    monitoring_url: Optional[StrictStr] = None
    message: Optional[StrictStr] = None
    resource_request: Optional[ResourceRequest] = None
    image_builder_resource_request: Optional[ResourceRequest] = None
    env_vars: Optional[List[EnvVar]] = None
    transformer: Optional[Transformer] = None
    logger: Optional[Logger] = None
    deployment_mode: Optional[DeploymentMode] = None
    autoscaling_policy: Optional[AutoscalingPolicy] = None
    protocol: Optional[Protocol] = None
    enable_model_observability: Optional[StrictBool] = None
    model_observability: Optional[ModelObservability] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    __properties: ClassVar[List[str]] = ["id", "version_id", "status", "url", "service_name", "environment_name", "environment", "monitoring_url", "message", "resource_request", "image_builder_resource_request", "env_vars", "transformer", "logger", "deployment_mode", "autoscaling_policy", "protocol", "enable_model_observability", "model_observability", "created_at", "updated_at"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of VersionEndpoint from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of environment
        if self.environment:
            _dict['environment'] = self.environment.to_dict()
        # override the default output from pydantic by calling `to_dict()` of resource_request
        if self.resource_request:
            _dict['resource_request'] = self.resource_request.to_dict()
        # override the default output from pydantic by calling `to_dict()` of image_builder_resource_request
        if self.image_builder_resource_request:
            _dict['image_builder_resource_request'] = self.image_builder_resource_request.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in env_vars (list)
        _items = []
        if self.env_vars:
            for _item in self.env_vars:
                if _item:
                    _items.append(_item.to_dict())
            _dict['env_vars'] = _items
        # override the default output from pydantic by calling `to_dict()` of transformer
        if self.transformer:
            _dict['transformer'] = self.transformer.to_dict()
        # override the default output from pydantic by calling `to_dict()` of logger
        if self.logger:
            _dict['logger'] = self.logger.to_dict()
        # override the default output from pydantic by calling `to_dict()` of autoscaling_policy
        if self.autoscaling_policy:
            _dict['autoscaling_policy'] = self.autoscaling_policy.to_dict()
        # override the default output from pydantic by calling `to_dict()` of model_observability
        if self.model_observability:
            _dict['model_observability'] = self.model_observability.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of VersionEndpoint from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "id": obj.get("id"),
            "version_id": obj.get("version_id"),
            "status": obj.get("status"),
            "url": obj.get("url"),
            "service_name": obj.get("service_name"),
            "environment_name": obj.get("environment_name"),
            "environment": Environment.from_dict(obj.get("environment")) if obj.get("environment") is not None else None,
            "monitoring_url": obj.get("monitoring_url"),
            "message": obj.get("message"),
            "resource_request": ResourceRequest.from_dict(obj.get("resource_request")) if obj.get("resource_request") is not None else None,
            "image_builder_resource_request": ResourceRequest.from_dict(obj.get("image_builder_resource_request")) if obj.get("image_builder_resource_request") is not None else None,
            "env_vars": [EnvVar.from_dict(_item) for _item in obj.get("env_vars")] if obj.get("env_vars") is not None else None,
            "transformer": Transformer.from_dict(obj.get("transformer")) if obj.get("transformer") is not None else None,
            "logger": Logger.from_dict(obj.get("logger")) if obj.get("logger") is not None else None,
            "deployment_mode": obj.get("deployment_mode"),
            "autoscaling_policy": AutoscalingPolicy.from_dict(obj.get("autoscaling_policy")) if obj.get("autoscaling_policy") is not None else None,
            "protocol": obj.get("protocol"),
            "enable_model_observability": obj.get("enable_model_observability"),
            "model_observability": ModelObservability.from_dict(obj.get("model_observability")) if obj.get("model_observability") is not None else None,
            "created_at": obj.get("created_at"),
            "updated_at": obj.get("updated_at")
        })
        return _obj





# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import client
from merlin.logger import Logger,LoggerMode,LoggerConfig,PredictionLoggerConfig


@pytest.mark.unit
def test_from_logger_response():
    logger_response = client.Logger(model=client.LoggerConfig(enabled=True, mode=client.LoggerMode.RESPONSE))
    result = Logger.from_logger_response(logger_response)
    expected_result = Logger(model=LoggerConfig(enabled=True, mode=LoggerMode.RESPONSE))
    assert result.model is not None
    assert result.model.enabled == expected_result.model.enabled
    assert result.model.mode == expected_result.model.mode
    assert result.transformer is None
    assert result.prediction is None

    logger_response = client.Logger(transformer=client.LoggerConfig(enabled=True, mode=client.LoggerMode.REQUEST))
    result = Logger.from_logger_response(logger_response)
    expected_result = Logger(transformer=LoggerConfig(enabled=True, mode=LoggerMode.REQUEST))
    assert result.transformer is not None
    assert result.transformer.enabled == expected_result.transformer.enabled
    assert result.transformer.mode == expected_result.transformer.mode
    assert result.model is None
    assert result.prediction is None

    logger_response = client.Logger(model=client.LoggerConfig(enabled=True, mode=client.LoggerMode.ALL),
                                    transformer=client.LoggerConfig(enabled=True, mode=client.LoggerMode.ALL))
    result = Logger.from_logger_response(logger_response)
    expected_result = Logger(model=LoggerConfig(enabled=True, mode=LoggerMode.ALL),
                             transformer=LoggerConfig(enabled=True, mode=LoggerMode.ALL))
    assert result.transformer is not None
    assert result.transformer.enabled == expected_result.transformer.enabled
    assert result.transformer.mode == expected_result.transformer.mode
    assert result.model is not None
    assert result.model.enabled == expected_result.model.enabled
    assert result.model.mode == expected_result.model.mode
    assert result.prediction is None

    result = Logger.from_logger_response(None)
    assert result.model is None
    assert result.transformer is None

    logger_response = client.Logger(model=client.LoggerConfig(enabled=True, mode=client.LoggerMode.ALL),
                                    transformer=client.LoggerConfig(enabled=True, mode=client.LoggerMode.ALL),
                                    prediction=client.PredictionLoggerConfig(enabled=True, raw_features_table="rawFeatures", entities_table="entities"))
    result = Logger.from_logger_response(logger_response)
    expected_result = Logger(model=LoggerConfig(enabled=True, mode=LoggerMode.ALL),
                             transformer=LoggerConfig(enabled=True, mode=LoggerMode.ALL),
                             prediction=PredictionLoggerConfig(enabled=True, raw_features_table="rawFeatures", entities_table="entities"))
    assert result.transformer is not None
    assert result.transformer.enabled == expected_result.transformer.enabled
    assert result.transformer.mode == expected_result.transformer.mode
    assert result.model is not None
    assert result.model.enabled == expected_result.model.enabled
    assert result.model.mode == expected_result.model.mode
    assert result.prediction is not None
    assert result.prediction.enabled == expected_result.prediction.enabled
    assert result.prediction.raw_features_table == expected_result.prediction.raw_features_table
    assert result.prediction.entities_table == expected_result.prediction.entities_table



@pytest.mark.unit
def test_to_logger_spec():
    logger = Logger(model=LoggerConfig(enabled=False, mode=LoggerMode.REQUEST))
    result = logger.to_logger_spec()
    expected_result = client.Logger(model=client.LoggerConfig(enabled=False, mode=client.LoggerMode.REQUEST))
    assert result.model is not None
    assert result.model.enabled == expected_result.model.enabled
    assert result.model.mode == expected_result.model.mode
    assert result.transformer is None
    assert result.prediction is None

    logger = Logger(transformer=LoggerConfig(enabled=True, mode=LoggerMode.RESPONSE))
    result = logger.to_logger_spec()
    expected_result = client.Logger(transformer=client.LoggerConfig(enabled=True, mode=client.LoggerMode.RESPONSE))
    assert result.transformer is not None
    assert result.transformer.enabled == expected_result.transformer.enabled
    assert result.transformer.mode == expected_result.transformer.mode
    assert result.model is None
    assert result.prediction is None


    logger = Logger(model=LoggerConfig(enabled=True, mode=LoggerMode.ALL),
                    transformer=LoggerConfig(enabled=True, mode=LoggerMode.ALL))
    result = logger.to_logger_spec()
    expected_result = client.Logger(model=client.LoggerConfig(enabled=True, mode=client.LoggerMode.ALL),
                                    transformer=client.LoggerConfig(enabled=True, mode=client.LoggerMode.ALL))
    assert result.transformer is not None
    assert result.transformer.enabled == expected_result.transformer.enabled
    assert result.transformer.mode == expected_result.transformer.mode
    assert result.model is not None
    assert result.model.enabled == expected_result.model.enabled
    assert result.model.mode == expected_result.model.mode
    assert result.prediction is None


    logger = Logger(model=LoggerConfig(enabled=True, mode=LoggerMode.ALL),
                    transformer=LoggerConfig(enabled=True, mode=LoggerMode.ALL),
                    prediction=PredictionLoggerConfig(enabled=True, raw_features_table="rawFeatures", entities_table="entities"))
    result = logger.to_logger_spec()
    expected_result = client.Logger(model=client.LoggerConfig(enabled=True, mode=client.LoggerMode.ALL),
                                    transformer=client.LoggerConfig(enabled=True, mode=client.LoggerMode.ALL),
                                    prediction=client.PredictionLoggerConfig(enabled=True, raw_features_table="rawFeatures", entities_table="entities"))
    assert result.transformer is not None
    assert result.transformer.enabled == expected_result.transformer.enabled
    assert result.transformer.mode == expected_result.transformer.mode
    assert result.model is not None
    assert result.model.enabled == expected_result.model.enabled
    assert result.model.mode == expected_result.model.mode
    assert result.prediction is not None
    assert result.prediction.enabled == expected_result.prediction.enabled
    assert result.prediction.raw_features_table == expected_result.prediction.raw_features_table
    assert result.prediction.entities_table == expected_result.prediction.entities_table

    logger = Logger()
    result = logger.to_logger_spec()
    assert result is None





# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import os
from merlin.transformer import StandardTransformer, TransformerType
from merlin.resource_request import ResourceRequest


@pytest.mark.unit
def test_feast_enricher():
    transformer_config_path = os.path.join("test/transformer", "feast_enricher.yaml")
    transformer = StandardTransformer(
        config_file=transformer_config_path, enabled=False
    )
    assert transformer.env_vars == {
        "STANDARD_TRANSFORMER_CONFIG": '{"transformerConfig": {"feast": [{"project": "merlin", "entities": [{"name": "merlin_test_driver_id", "valueType": "STRING", "jsonPath": "$.driver_id"}], "features": [{"name": "merlin_test_driver_features:test_int32", "valueType": "INT32", "defaultValue": "0"}, {"name": "merlin_test_driver_features:test_float", "valueType": "FLOAT", "defaultValue": "0.0"}, {"name": "merlin_test_driver_features:test_double", "valueType": "DOUBLE", "defaultValue": "0.0"}, {"name": "merlin_test_driver_features:test_string", "valueType": "STRING", "defaultValue": ""}]}]}}'
    }
    assert not transformer.enabled
    assert transformer.command is None
    assert transformer.args is None
    assert transformer.transformer_type == TransformerType.STANDARD_TRANSFORMER


def test_feast_enricher_with_env_vars():
    transformer_config_path = os.path.join("test/transformer", "feast_enricher.yaml")
    resource = ResourceRequest(
        min_replica=1, max_replica=2, cpu_request="100m", memory_request="128Mi"
    )
    transformer = StandardTransformer(
        config_file=transformer_config_path,
        enabled=True,
        resource_request=resource,
        env_vars={"MODEL_URL": "http://model.default"},
    )
    assert transformer.env_vars == {
        "MODEL_URL": "http://model.default",
        "STANDARD_TRANSFORMER_CONFIG": '{"transformerConfig": {"feast": [{"project": "merlin", "entities": [{"name": "merlin_test_driver_id", "valueType": "STRING", "jsonPath": "$.driver_id"}], "features": [{"name": "merlin_test_driver_features:test_int32", "valueType": "INT32", "defaultValue": "0"}, {"name": "merlin_test_driver_features:test_float", "valueType": "FLOAT", "defaultValue": "0.0"}, {"name": "merlin_test_driver_features:test_double", "valueType": "DOUBLE", "defaultValue": "0.0"}, {"name": "merlin_test_driver_features:test_string", "valueType": "STRING", "defaultValue": ""}]}]}}',
    }
    assert transformer.enabled
    assert transformer.command is None
    assert transformer.args is None
    assert transformer.resource_request.min_replica == 1
    assert transformer.resource_request.max_replica == 2
    assert transformer.resource_request.cpu_request == "100m"
    assert transformer.resource_request.memory_request == "128Mi"
    assert transformer.transformer_type == TransformerType.STANDARD_TRANSFORMER



# coding: utf-8

# flake8: noqa

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


__version__ = "1.0.0"

# import apis into sdk package
from client.api.alert_api import AlertApi
from client.api.endpoint_api import EndpointApi
from client.api.environment_api import EnvironmentApi
from client.api.log_api import LogApi
from client.api.model_endpoints_api import ModelEndpointsApi
from client.api.model_schema_api import ModelSchemaApi
from client.api.models_api import ModelsApi
from client.api.prediction_jobs_api import PredictionJobsApi
from client.api.project_api import ProjectApi
from client.api.secret_api import SecretApi
from client.api.standard_transformer_api import StandardTransformerApi
from client.api.version_api import VersionApi
from client.api.version_image_api import VersionImageApi

# import ApiClient
from client.api_response import ApiResponse
from client.api_client import ApiClient
from client.configuration import Configuration
from client.exceptions import OpenApiException
from client.exceptions import ApiTypeError
from client.exceptions import ApiValueError
from client.exceptions import ApiKeyError
from client.exceptions import ApiAttributeError
from client.exceptions import ApiException

# import models into sdk package
from client.models.alert_condition_metric_type import AlertConditionMetricType
from client.models.alert_condition_severity import AlertConditionSeverity
from client.models.autoscaling_policy import AutoscalingPolicy
from client.models.binary_classification_output import BinaryClassificationOutput
from client.models.build_image_options import BuildImageOptions
from client.models.config import Config
from client.models.container import Container
from client.models.custom_predictor import CustomPredictor
from client.models.deployment_mode import DeploymentMode
from client.models.endpoint_status import EndpointStatus
from client.models.env_var import EnvVar
from client.models.environment import Environment
from client.models.file_format import FileFormat
from client.models.gpu_config import GPUConfig
from client.models.gpu_toleration import GPUToleration
from client.models.ground_truth_job import GroundTruthJob
from client.models.ground_truth_source import GroundTruthSource
from client.models.image_building_job_state import ImageBuildingJobState
from client.models.image_building_job_status import ImageBuildingJobStatus
from client.models.label import Label
from client.models.list_jobs_paginated_response import ListJobsPaginatedResponse
from client.models.logger import Logger
from client.models.logger_config import LoggerConfig
from client.models.logger_mode import LoggerMode
from client.models.metrics_type import MetricsType
from client.models.mock_response import MockResponse
from client.models.model import Model
from client.models.model_endpoint import ModelEndpoint
from client.models.model_endpoint_alert import ModelEndpointAlert
from client.models.model_endpoint_alert_condition import ModelEndpointAlertCondition
from client.models.model_endpoint_rule import ModelEndpointRule
from client.models.model_endpoint_rule_destination import ModelEndpointRuleDestination
from client.models.model_observability import ModelObservability
from client.models.model_prediction_config import ModelPredictionConfig
from client.models.model_prediction_output import ModelPredictionOutput
from client.models.model_prediction_output_class import ModelPredictionOutputClass
from client.models.model_schema import ModelSchema
from client.models.operation_tracing import OperationTracing
from client.models.paging import Paging
from client.models.pipeline_tracing import PipelineTracing
from client.models.prediction_job import PredictionJob
from client.models.prediction_job_config import PredictionJobConfig
from client.models.prediction_job_config_bigquery_sink import PredictionJobConfigBigquerySink
from client.models.prediction_job_config_bigquery_source import PredictionJobConfigBigquerySource
from client.models.prediction_job_config_gcs_sink import PredictionJobConfigGcsSink
from client.models.prediction_job_config_gcs_source import PredictionJobConfigGcsSource
from client.models.prediction_job_config_model import PredictionJobConfigModel
from client.models.prediction_job_config_model_result import PredictionJobConfigModelResult
from client.models.prediction_job_resource_request import PredictionJobResourceRequest
from client.models.prediction_log_ingestion_resource_request import PredictionLogIngestionResourceRequest
from client.models.prediction_logger_config import PredictionLoggerConfig
from client.models.project import Project
from client.models.protocol import Protocol
from client.models.ranking_output import RankingOutput
from client.models.regression_output import RegressionOutput
from client.models.resource_request import ResourceRequest
from client.models.result_type import ResultType
from client.models.save_mode import SaveMode
from client.models.schema_spec import SchemaSpec
from client.models.secret import Secret
from client.models.standard_transformer_simulation_request import StandardTransformerSimulationRequest
from client.models.standard_transformer_simulation_response import StandardTransformerSimulationResponse
from client.models.transformer import Transformer
from client.models.value_type import ValueType
from client.models.version import Version
from client.models.version_endpoint import VersionEndpoint
from client.models.version_image import VersionImage



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictStr
from client.models.prediction_job_config_bigquery_sink import PredictionJobConfigBigquerySink
from client.models.prediction_job_config_bigquery_source import PredictionJobConfigBigquerySource
from client.models.prediction_job_config_gcs_sink import PredictionJobConfigGcsSink
from client.models.prediction_job_config_gcs_source import PredictionJobConfigGcsSource
from client.models.prediction_job_config_model import PredictionJobConfigModel
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class PredictionJobConfig(BaseModel):
    """
    PredictionJobConfig
    """ # noqa: E501
    version: Optional[StrictStr] = None
    kind: Optional[StrictStr] = None
    name: Optional[StrictStr] = None
    bigquery_source: Optional[PredictionJobConfigBigquerySource] = None
    gcs_source: Optional[PredictionJobConfigGcsSource] = None
    model: Optional[PredictionJobConfigModel] = None
    bigquery_sink: Optional[PredictionJobConfigBigquerySink] = None
    gcs_sink: Optional[PredictionJobConfigGcsSink] = None
    __properties: ClassVar[List[str]] = ["version", "kind", "name", "bigquery_source", "gcs_source", "model", "bigquery_sink", "gcs_sink"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of PredictionJobConfig from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of bigquery_source
        if self.bigquery_source:
            _dict['bigquery_source'] = self.bigquery_source.to_dict()
        # override the default output from pydantic by calling `to_dict()` of gcs_source
        if self.gcs_source:
            _dict['gcs_source'] = self.gcs_source.to_dict()
        # override the default output from pydantic by calling `to_dict()` of model
        if self.model:
            _dict['model'] = self.model.to_dict()
        # override the default output from pydantic by calling `to_dict()` of bigquery_sink
        if self.bigquery_sink:
            _dict['bigquery_sink'] = self.bigquery_sink.to_dict()
        # override the default output from pydantic by calling `to_dict()` of gcs_sink
        if self.gcs_sink:
            _dict['gcs_sink'] = self.gcs_sink.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of PredictionJobConfig from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "version": obj.get("version"),
            "kind": obj.get("kind"),
            "name": obj.get("name"),
            "bigquery_source": PredictionJobConfigBigquerySource.from_dict(obj.get("bigquery_source")) if obj.get("bigquery_source") is not None else None,
            "gcs_source": PredictionJobConfigGcsSource.from_dict(obj.get("gcs_source")) if obj.get("gcs_source") is not None else None,
            "model": PredictionJobConfigModel.from_dict(obj.get("model")) if obj.get("model") is not None else None,
            "bigquery_sink": PredictionJobConfigBigquerySink.from_dict(obj.get("bigquery_sink")) if obj.get("bigquery_sink") is not None else None,
            "gcs_sink": PredictionJobConfigGcsSink.from_dict(obj.get("gcs_sink")) if obj.get("gcs_sink") is not None else None
        })
        return _obj


